#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Typed

from enaml.widgets.frame import ProxyFrame

from .QtWidgets import QFrame

from .qt_constraints_widget import QtConstraintsWidget


STYLE = {
    'box': QFrame.Box,
    'panel': QFrame.Panel,
    'styled_panel': QFrame.StyledPanel,
}


LINE_STYLE = {
    'plain': QFrame.Plain,
    'sunken': QFrame.Sunken,
    'raised': QFrame.Raised,
}


class QtFrame(QtConstraintsWidget, ProxyFrame):
    """ A Qt implementation of an Enaml ProxyFrame.

    """
    #: A reference to the toolkit widget created by the proxy.
    widget = Typed(QFrame)

    #--------------------------------------------------------------------------
    # Initialization API
    #--------------------------------------------------------------------------
    def create_widget(self):
        """ Creates the QContainer widget.

        """
        self.widget = QFrame(self.parent_widget())

    def init_widget(self):
        """ Initialize the widget.

        """
        super(QtFrame, self).init_widget()
        self.set_border(self.declaration.border)

    #--------------------------------------------------------------------------
    # ProxyFrame API
    #--------------------------------------------------------------------------
    def set_border(self, border):
        """ Set the border for the widget.

        """
        widget = self.widget
        if border is None:
            widget.setFrameShape(QFrame.NoFrame)
            return
        widget.setFrameShape(STYLE[border.style])
        widget.setFrameShadow(LINE_STYLE[border.line_style])
        widget.setLineWidth(border.line_width)
        widget.setMidLineWidth(border.midline_width)



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Bool, Typed, ForwardTyped, observe

from enaml.core.declarative import d_

from .abstract_button import AbstractButton, ProxyAbstractButton
from .menu import Menu


class ProxyPushButton(ProxyAbstractButton):
    """ The abstract definition of a proxy PushButton object.

    """
    #: A reference to the PushButton declaration.
    declaration = ForwardTyped(lambda: PushButton)

    def set_default(self, default):
        raise NotImplementedError


class PushButton(AbstractButton):
    """ A button control represented by a standard push button widget.

    """
    #: Whether this button is the default action button in a dialog.
    default = d_(Bool(False))

    #: A reference to the ProxyPushButton object.
    proxy = Typed(ProxyPushButton)

    def menu(self):
        """ Get the menu defined for the PushButton, if any.

        """
        for child in reversed(self.children):
            if isinstance(child, Menu):
                return child

    #--------------------------------------------------------------------------
    # Observers
    #--------------------------------------------------------------------------
    @observe('default')
    def _update_proxy(self, change):
        """ Send the member state change to the proxy.

        """
        # The superclass implementation is sufficient
        super(PushButton, self)._update_proxy(change)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2020, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from contextlib import contextmanager

import bytecode as bc
from atom.api import Atom, Bool, Int, List, Str

from ..compat import POS_ONLY_ARGS, PY38, PY39


class CodeGenerator(Atom):
    """ A class for generating bytecode operations.

    This class matches enaml needs and does not support freevars or cellvars.

    """
    #: The arguments for the code.
    args = List()

    #: Number of positional only arguments
    posonlyargs = Int()

    #: Number of keyword only arguments
    kwonlyargs = Int()

    #: Whether the code takes variadic args.
    varargs = Bool(False)

    #: Whether the code takes variadic kwargs.
    varkwargs = Bool(False)

    #: Whether the code object should get new locals.
    newlocals = Bool(False)

    #: The name for the code object.
    name = Str()

    #: The full name of the file which is being compiled.
    filename = Str()

    #: The first line number for the code object.
    firstlineno = Int(1)

    #: The docstring for the code object.
    docstring = Str()

    #: The list of generated code operations.
    code_ops = List()

    def to_code(self):
        """ Create a Python code object from the current code ops.

        """
        bc_code = bc.Bytecode(self.code_ops)
        bc_code.argnames = self.args
        # The number of positional or keyword args correspond to all args minus:
        # - the positionals only
        # - the keywords only
        # - the variadic positional
        # - the variadic keyword
        bc_code.argcount = (len(self.args) -
                            self.kwonlyargs -
                            self.posonlyargs -
                            self.varargs -
                            self.varkwargs)
        if POS_ONLY_ARGS:
            bc_code.posonlyargcount = self.posonlyargs
        bc_code.kwonlyargcount = self.kwonlyargs


        for name in ("name", "filename", "firstlineno", "docstring"):
            setattr(bc_code, name, getattr(self, name))

        # Set flags appropriately and update flags based on the instructions
        for setting, flag in zip((self.varargs, self.varkwargs, self.newlocals),
                                 (bc.CompilerFlags.VARARGS,
                                  bc.CompilerFlags.VARKEYWORDS,
                                  bc.CompilerFlags.NEWLOCALS)
                                 ):
            # Set the flag
            if setting:
                bc_code.flags |= flag
            # Unset the flag if it was set
            else:
                bc_code.flags ^= (bc_code.flags & flag)
        bc_code.update_flags()

        return bc_code.to_code()

    def set_lineno(self, lineno):
        """ Set the current line number in the code.

        """
        self.code_ops.append(                           # TOS
            bc.SetLineno(lineno),                       # TOS
        )

    def load_global(self, name):
        """ Load a global variable onto the TOS.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("LOAD_GLOBAL", name),              # TOS -> value
        )

    def load_name(self, name):
        """ Load a named variable onto the TOS.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("LOAD_NAME", name),                # TOS -> value
        )

    def load_fast(self, name):
        """ Load a fast local variable onto the TOS.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("LOAD_FAST", name),                # TOS -> value
        )

    def load_const(self, const):
        """ Load a const value onto the TOS.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("LOAD_CONST", const),              # TOS -> value
        )

    def load_attr(self, name):
        """ Load an attribute from the object on TOS.

        """
        self.code_ops.append(                           # TOS -> obj
            bc.Instr("LOAD_ATTR", name),                # TOS -> value
        )

    def store_global(self, name):
        """ Store the TOS as a global.

        """
        self.code_ops.append(                           # TOS -> value
            bc.Instr("STORE_GLOBAL", name),             # TOS
        )

    def store_name(self, name):
        """ Store the TOS under name.

        """
        self.code_ops.append(                           # TOS -> value
            bc.Instr("STORE_NAME", name),             # TOS
        )

    def store_fast(self, name):
        """ Store the TOS as a fast local.

        """
        self.code_ops.append(                           # TOS -> value
            bc.Instr("STORE_FAST", name),               # TOS
        )

    def store_attr(self, name):
        """ Store the value at 2nd as an attr on 1st.

        """
        self.code_ops.append(                           # TOS -> val -> obj
            bc.Instr("STORE_ATTR", name),               # TOS
        )

    def delete_global(self, name):
        """ Delete a named global variable.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("DELETE_GLOBAL", name),            # TOS
        )

    def delete_fast(self, name):
        """ Delete a named fast local variable.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("DELETE_FAST", name),                     # TOS
        )

    def return_value(self):
        """ Return the value from the TOS.

        """
        self.code_ops.append(                           # TOS -> value
            bc.Instr("RETURN_VALUE"),                   # TOS
        )

    def binary_subscr(self):
        """ Subscript the #2 item with the TOS.

        """
        self.code_ops.append(                           # TOS -> obj -> idx
            bc.Instr("BINARY_SUBSCR"),                  # TOS -> value
        )

    def binary_multiply(self):
        """ Multiple the 2 items on the TOS.

        """
        self.code_ops.append(                           # TOS -> val_1 -> val_2
            bc.Instr("BINARY_MULTIPLY"),                # TOS -> retval
        )

    def binary_add(self):
        """ Multiple the 2 items on the TOS.

        """
        self.code_ops.append(                           # TOS -> val_1 -> val_2
            bc.Instr("BINARY_ADD"),                     # TOS -> retval
        )

    def dup_top(self):
        """ Duplicate the value on the TOS.

        """
        self.code_ops.append(                           # TOS -> value
            bc.Instr("DUP_TOP"),                        # TOS -> value -> value
        )

    def build_map(self, n=0):
        """ Build a map and store it onto the TOS.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("BUILD_MAP", n),                   # TOS -> map
        )

    def build_tuple(self, n=0):
        """ Build a tuple from items on the TOS.

        """
        if n == 0:
            self.code_ops.append(                       # TOS
                bc.Instr("LOAD_CONST", ()),             # TOS -> tuple
            )
        else:
            self.code_ops.append(                       # TOS
                bc.Instr("BUILD_TUPLE", n),             # TOS -> tuple
            )

    def build_list(self, n=0):
        """ Build a list from items on the TOS.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("BUILD_LIST", n),                  # TOS -> list
        )

    def add_map(self):
        """ Store the key/value pair on the TOS into the map at 3rd pos.

        """
        # WARNING in Python 3.8 the order is           # TOS -> map -> key -> value
        self.code_ops.append(                          # TOS -> map -> value -> key
            bc.Instr("MAP_ADD", 1),
        )

    def store_subscr(self):
        """ Store the index/value pair on the TOS into the 3rd item.

        """
        self.code_ops.append(                           # TOS -> value -> obj -> index
            bc.Instr("STORE_SUBSCR"),                   # TOS
        )

    def load_build_class(self):
        """ Build a class from the top 3 stack items.

        """
        self.code_ops.append(                           # TOS
            bc.Instr("LOAD_BUILD_CLASS"),               # TOS -> builtins.__build_class__
        )

    def make_function(self, n_defaults=0):
        """ Make a function from a code object on the TOS.

        """
        self.code_ops.append(                           # TOS -> qual_name -> code -> defaults
            bc.Instr("MAKE_FUNCTION", n_defaults),      # TOS -> func
        )

    def call_function(self, n_args=0, n_kwds=0):
        """ Call a function on the TOS with the given args and kwargs.

        """
        if n_kwds:
            # kwargs_name should be a tuple listing the keyword
            # arguments names
            # TOS -> func -> args -> kwargs -> kwargs_names
            op, arg = "CALL_FUNCTION_KW", n_args+n_kwds
        else:
            op, arg = "CALL_FUNCTION", n_args

        self.code_ops.append(bc.Instr(op, arg))          # TOS -> retval

    def call_function_var(self, n_args=0, n_kwds=0):
        """ Call a variadic function on the TOS with the given args and kwargs.

        """
        # Under Python 3.6+ positional arguments should always be stored
        # in a tuple and keywords in a mapping.
        argspec = 1 if n_kwds else 0

        self.code_ops.append(                           # TOS -> func -> args -> kwargs -> varargs
            bc.Instr("CALL_FUNCTION_EX", argspec),      # TOS -> retval
        )

    def pop_top(self):
        """ Pop the value from the TOS.

        """
        self.code_ops.append(                           # TOS -> value
            bc.Instr("POP_TOP"),                        # TOS
        )

    def rot_two(self):
        """ Rotate the two values on the TOS.

        """
        self.code_ops.append(                           # TOS -> val_1 -> val_2
            bc.Instr("ROT_TWO"),                        # TOS -> val_2 -> val_1
        )

    def rot_three(self):
        """ Rotate the three values on the TOS.

        """
        self.code_ops.append(                           # TOS -> val_1 -> val_2 -> val_3
            bc.Instr("ROT_THREE"),                      # TOS -> val_3 -> val_1 -> val_2
        )

    def unpack_sequence(self, n):
        """ Unpack the sequence on the TOS.

        """
        self.code_ops.append(                           # TOS -> obj
            bc.Instr("UNPACK_SEQUENCE", n),             # TOS -> val_n -> val_2 -> val_1
        )

    @contextmanager
    def try_squash_raise(self):
        """ A context manager for squashing tracebacks.

        The code written during this context will be wrapped so that
        any exception raised will appear to have been generated from
        the code, rather than any function called by the code.

        """
        exc_label = bc.Label()
        end_label = bc.Label()
        op_code = "SETUP_FINALLY" if PY38 else "SETUP_EXCEPT"
        self.code_ops.append(
            bc.Instr(op_code, exc_label),              # TOS
        )
        yield
        # exc is only the exception type which can be used for matching
        # val is the exception value, raising it directly preserve the traceback
        # tb is the traceback and is of little interest
        # We reset the traceback to None to make it appear as if the code
        # raised the exception instead of a function called by it
        ops = [                                       # TOS
            bc.Instr("POP_BLOCK"),                    # TOS
            bc.Instr("JUMP_FORWARD", end_label),      # TOS
            exc_label,                                # TOS -> tb -> val -> exc
            bc.Instr("POP_TOP"),                      # TOS -> tb -> val
            bc.Instr("ROT_TWO"),                      # TOS -> val -> tb
            bc.Instr("POP_TOP"),                      # TOS -> val
            bc.Instr("DUP_TOP"),                      # TOS -> val -> val
            bc.Instr("LOAD_CONST", None),             # TOS -> val -> val -> None
            bc.Instr("ROT_TWO"),                      # TOS -> val -> None -> val
            bc.Instr("STORE_ATTR", "__traceback__"),  # TOS -> val
            bc.Instr("RAISE_VARARGS", 1),             # TOS
            bc.Instr("POP_EXCEPT"),
            bc.Instr("JUMP_FORWARD", end_label),      # TOS
            end_label,                                # TOS
        ]
        if not PY39:
            ops.insert(-1, bc.Instr("END_FINALLY"))
        self.code_ops.extend(ops)


    @contextmanager
    def for_loop(self, iter_var, fast_var=True):
        """ A context manager for creating for-loops.

        Parameters
        ----------
        iter_var : str
            The name of the loop iter variable.

        fast_var : bool, optional
            Whether the iter_var lives in fast locals. The default is
            True. If False, the iter_var is loaded from globals.

        """
        start_label = bc.Label()
        jump_label = bc.Label()
        # Unused under Python 3.8+ since the compiler handle the blocks
        # automatically
        end_label = bc.Label()
        load_op = "LOAD_FAST" if fast_var else "LOAD_GLOBAL"
        if PY38:
            self.code_ops.append(bc.Instr("SETUP_LOOP", end_label),)
        self.code_ops.extend([
            bc.Instr(load_op, iter_var),
            bc.Instr("GET_ITER"),
            start_label,
            bc.Instr("FOR_ITER", jump_label),
        ])
        yield
        self.code_ops.extend([
            bc.Instr("JUMP_ABSOLUTE", start_label),
            jump_label,
        ])
        if PY38:
            self.code_ops.extend(
                bc.Instr("POP_BLOCK"),
                end_label,
            )

    def insert_python_block(self, pydata, trim=True):
        """ Insert the compiled code for a Python Module ast or string.

        """
        code = compile(pydata, self.filename, mode='exec')
        bc_code = bc.Bytecode.from_code(code)
        if trim:  # skip  ReturnValue
            bc_code = bc_code[:-1]
        self.code_ops.extend(bc_code)

    def insert_python_expr(self, pydata, trim=True):
        """ Insert the compiled code for a Python Expression ast or string.

        """
        code = compile(pydata, self.filename, mode='eval')
        bc_code = bc.Bytecode.from_code(code)
        if trim:  # skip ReturnValue
            bc_code = bc_code[:-1]
        self.code_ops.extend(bc_code)

    def rewrite_to_fast_locals(self, local_names):
        """ Rewrite the locals to be loaded from fast locals.

        Given a set of available local names, this method will rewrite
        the current code ops, replaces every instance of a *_NAME opcode
        with a *_FAST or *_GLOBAL depending on whether or not the name
        exists in local_names or was written via STORE_NAME. This method
        is useful to convert the code so it can be used as a function.

        Parameters
        ----------
        local_names : set
            The set of available locals for the code.

        Returns
        -------
        result : list
            The list of names which must be provided as arguments.

        """
        arg_names = []
        stored_names = set()
        code_ops = self.code_ops
        for idx, instr in enumerate(code_ops):
            if instr.name == "STORE_NAME":
                stored_names.add(instr.arg)
                instr.name = "STORE_FAST"
        for idx, instr in enumerate(code_ops):
            i_name = instr.name
            if i_name == "LOAD_NAME":
                i_arg = instr.arg
                if i_arg in local_names:
                    op = "LOAD_FAST"
                    arg_names.append(i_arg)
                elif i_arg in stored_names:
                    op = "LOAD_FAST"
                else:
                    op = "LOAD_GLOBAL"
                instr.name = op
            elif i_name == "DELETE_NAME":
                if instr.arg in stored_names:
                    op = "DELETE_FAST"
                else:
                    op = "DELETE_GLOBAL"
                instr.name = op
        self.args = arg_names
        self.newlocals = True
        return arg_names



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Typed

from enaml.widgets.mdi_area import ProxyMdiArea

from .QtWidgets import QMdiArea

from .qt_constraints_widget import QtConstraintsWidget
from .qt_mdi_window import QtMdiWindow


class QtMdiArea(QtConstraintsWidget, ProxyMdiArea):
    """ A Qt implementation of an Enaml ProxyMdiArea.

    """
    #: A reference to the widget created by the proxy.
    widget = Typed(QMdiArea)

    #--------------------------------------------------------------------------
    # Initialization API
    #--------------------------------------------------------------------------
    def create_widget(self):
        """ Create the underlying QMdiArea widget.

        """
        self.widget = QMdiArea(self.parent_widget())

    def init_layout(self):
        """ Initialize the layout for the underlying control.

        """
        super(QtMdiArea, self).init_layout()
        widget = self.widget
        for window in self.mdi_windows():
            widget.addSubWindow(window)
        widget.subWindowActivated.connect(self.on_subwindow_activated)

    #--------------------------------------------------------------------------
    # Utility Methods
    #--------------------------------------------------------------------------
    def mdi_windows(self):
        """ Get the mdi windows defined for the area.

        """
        for window in self.declaration.mdi_windows():
            widget = window.proxy.widget
            if widget:
                yield widget

    def tile_mdi_windows(self):
        """ Tile the subwindows.

        Notes
        -----
        For the time being the ordering is the insertion order (Qt default). In
        the future, a way to modify the order may be added.

        """
        if self.widget:
            self.widget.tileSubWindows()

    def cascade_mdi_windows(self):
        """ Cascade the subwindows.

        Notes
        -----
        For the time being the ordering is the insertion order (Qt default). In
        the future, a way to modify the order may be added.

        """
        if self.widget:
            self.widget.cascadeSubWindows()

    #--------------------------------------------------------------------------
    # Signal Handlers
    #--------------------------------------------------------------------------
    def on_subwindow_activated(self, window):
        """ The handler for the 'subWindowActivated' signal.

        """
        # On OSX there is painting bug where a subwindow is not repainted
        # properly when it is activated. This handler ensures an update.
        if window:
            window.update()

    #--------------------------------------------------------------------------
    # Child Events
    #--------------------------------------------------------------------------
    def child_added(self, child):
        """ Handle the child added event for a QtMdiArea.

        """
        # The size hint of a QMdiArea is typically quite large and the
        # size hint constraints are usually ignored. There is no need
        # to notify of a change in size hint here.
        super(QtMdiArea, self).child_added(child)
        if isinstance(child, QtMdiWindow):
            self.widget.addSubWindow(child.widget)

    def child_removed(self, child):
        """ Handle the child removed event for a QtMdiArea.

        """
        if isinstance(child, QtMdiWindow):
            self.widget.removeSubWindow(child.widget)
        else:
            super(QtMdiArea, self).child_removed(child)



#------------------------------------------------------------------------------
# Copyright (c) 2014, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Atom, IntEnum, Typed, Coerced, Tuple, Int

from .application import Application
from .image import Image
from .mime_data import MimeData


class DropAction(IntEnum):
    """ An enum defining the possible drop actions.

    """
    #: The action is ignored.
    Ignore = 0x0

    #: The data is copied to the target.
    Copy = 0x1

    #: The data is moved from the source to the target.
    Move = 0x2

    #: Create a link from the source to the target.
    Link = 0x4


def mime_data_factory():
    """ Create a new MimeData object for a drag operation.

    Returns
    -------
    result : MimeData
        A toolkit specific mime data object.

    """
    return Application.instance().create_mime_data()


class DragData(Atom):
    """ An object which initialize the data for a drag operation.

    """
    #: The mime data to use for the drag operation. This is created
    #: automatically, but can be reassigned by the user if necessary.
    mime_data = Typed(MimeData, factory=mime_data_factory)

    #: The default drop action for the drag data. If not provided,
    #: the toolkit will choose a suitable default from among the
    #: supported action.
    default_drop_action = Coerced(DropAction, (DropAction.Ignore,))

    #: The supported drop actions of the drag data. This is an OR'd
    #: combination of the available DropAction flags.
    supported_actions = Coerced(DropAction.Flags, (DropAction.Move,))

    #: The image to use for the drag. If this is not provided, the
    #: toolkit will choose a suitable default value.
    image = Typed(Image)

    #: The x,y position the drag image appears under the cursor.
    #: If not provided, this is the last position of the cursor in the widget.
    hotspot = Tuple(item=Int())


class DropEvent(Atom):
    """ An abstract class for defining a drag event.

    Concrete implementations of this class will be created by a
    toolkit backend and passed to the relevant frontend handlers.

    Instances of this class will never be created by the user.

    """
    def pos(self):
        """ Get the current mouse position of the operation.

        Returns
        -------
        result : Pos
            The mouse position of the operation in widget coordinates.

        """
        raise NotImplementedError

    def mime_data(self):
        """ Get the mime data contained in the drag operation.

        Returns
        -------
        result : MimeData
            The mime data contained in the drag operation.

        """
        raise NotImplementedError

    def drop_action(self):
        """ Get the action to be performed by the drop target.

        Returns
        -------
        result : DropAction
            A drop action enum value.

        """
        raise NotImplementedError

    def possible_actions(self):
        """ Get the OR'd combination of possible drop actions.

        Returns
        -------
        result : DropAction.Flags
            The combination of possible drop actions.

        """
        raise NotImplementedError

    def proposed_action(self):
        """ Get the action proposed to be taken by the drop target.

        Returns
        -------
        result : DropAction
            The proposed action for the drop target.

        """
        raise NotImplementedError

    def accept_proposed_action(self):
        """ Accept the event using the proposed drop action.

        """
        raise NotImplementedError

    def set_drop_action(self, action):
        """ Set the drop action to one of the possible actions.

        Parameters
        ----------
        action : DropAction
            The drop action to be performed by the target.

        """
        raise NotImplementedError

    def is_accepted(self):
        """ Test whether the event has been accepted.

        Returns
        -------
        result : bool
            True if the event is accepted, False otherwise.

        """
        raise NotImplementedError

    def set_accepted(self, accepted):
        """ Set the accepted state of the event.

        Parameters
        ----------
        accepted : bool
            The target accepted state of the event.

        """
        raise NotImplementedError

    def accept(self):
        """ Accept the current drop event action.

        """
        raise NotImplementedError

    def ignore(self):
        """ Ignore the current drop event action.

        """
        raise NotImplementedError
